// Code generated by mockery v1.0.0. DO NOT EDIT.

package gocbcore

import mock "github.com/stretchr/testify/mock"

// mockHttpComponentInterface is an autogenerated mock type for the httpComponentInterface type
type mockHttpComponentInterface struct {
	mock.Mock
}

// DoInternalHTTPRequest provides a mock function with given fields: req, skipConfigCheck
func (_m *mockHttpComponentInterface) DoInternalHTTPRequest(req *httpRequest, skipConfigCheck bool) (*HTTPResponse, error) {
	ret := _m.Called(req, skipConfigCheck)

	var r0 *HTTPResponse
	if rf, ok := ret.Get(0).(func(*httpRequest, bool) *HTTPResponse); ok {
		r0 = rf(req, skipConfigCheck)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(*HTTPResponse)
		}
	}

	var r1 error
	if rf, ok := ret.Get(1).(func(*httpRequest, bool) error); ok {
		r1 = rf(req, skipConfigCheck)
	} else {
		r1 = ret.Error(1)
	}

	return r0, r1
}
